import typing
from synchronicity import Synchronizer


synchronizer = Synchronizer()


async def foo() -> typing.AsyncGenerator[int, None]:
    yield 1


foo = synchronizer.wrap(foo, target_module="my_library")


import asyncio

from synchronicity import Synchronizer
from synchronicity.async_utils import Runner


async def run():
    try:
        while True:
            await asyncio.sleep(0.2)
            print("running")

    except asyncio.CancelledError:
        print("cancelled")
        await asyncio.sleep(0.1)
        print("handled cancellation")
        raise
    finally:
        await asyncio.sleep(0.1)
        print("exit async")


s = Synchronizer()

blocking_run = s.create_blocking(run)


try:
    with Runner() as runner:
        runner.run(blocking_run.aio())
except KeyboardInterrupt:
    print("keyboard interrupt")



import asyncio
import inspect


class Callback:
    """A callback is when synchronized call needs to call outside functions passed into it.

    Currently only supports non-generator functions."""

    def __init__(self, synchronizer, f):
        self._synchronizer = synchronizer
        self._f = f

    def _invoke(self, args, kwargs):
        # This runs on a separate thread
        res = self._f(*args, **kwargs)
        if inspect.iscoroutine(res):
            try:
                loop = asyncio.new_event_loop()
                return loop.run_until_complete(res)
            finally:
                loop.close()
        elif inspect.isasyncgen(res):
            raise RuntimeError("Async generators are not supported")
        elif inspect.isgenerator(res):
            raise RuntimeError("Generators are not supported")
        else:
            return res

    async def __call__(self, *args, **kwargs):
        # This translates the opposite way from the code in the synchronizer
        args = self._synchronizer._translate_out(args)
        kwargs = self._synchronizer._translate_out(kwargs)

        # This function may be blocking, so we need to run it on a thread
        loop = asyncio.get_event_loop()
        res = await loop.run_in_executor(None, self._invoke, args, kwargs)

        # Now, we need to translate the result _in_
        return self._synchronizer._translate_in(res)



import asyncio
import time


def test_nowrap(synchronizer):
    @synchronizer.create_blocking
    class MyClass:
        async def f(self, x):
            await asyncio.sleep(0.2)
            return x**2

        @synchronizer.nowrap
        def g(self, x):
            # This runs on the wrapped class
            return self.f(x) * x  # calls the blocking function

    my_obj = MyClass()

    t0 = time.time()
    assert my_obj.f(111) == 12321
    assert 0.15 < time.time() - t0 < 0.25

    t0 = time.time()
    assert my_obj.g(111) == 1367631
    assert 0.15 < time.time() - t0 < 0.25



def test_translate(synchronizer):
    class Foo:
        pass

    class FooProvider:
        def __init__(self, foo=None):
            if foo is not None:
                assert type(foo) is Foo
                self.foo = foo
            else:
                self.foo = Foo()

        def get(self):
            return self.foo

        def get2(self):
            return [self.foo, self.foo]

        @property
        def pget(self):
            return self.foo

        def set(self, foo):
            assert type(foo) is Foo
            self.foo = foo

        @classmethod
        def cls_in(cls):
            assert cls == FooProvider

        @classmethod
        def cls_out(cls):
            return FooProvider

    BlockingFoo = synchronizer.create_blocking(Foo)
    assert BlockingFoo.__name__ == "BlockingFoo"
    BlockingFooProvider = synchronizer.create_blocking(FooProvider)
    assert BlockingFooProvider.__name__ == "BlockingFooProvider"
    foo_provider = BlockingFooProvider()

    # Make sure two instances translated out are the same
    foo1 = foo_provider.get()
    foo2 = foo_provider.get()
    assert foo1 == foo2

    # Make sure we can return a list
    foos = foo_provider.get2()
    assert foos == [foo1, foo2]

    # Make sure properties work
    foo = foo_provider.pget
    assert isinstance(foo, BlockingFoo)

    # Translate an object in and then back out, make sure it's the same
    foo = BlockingFoo()
    assert type(foo) is BlockingFoo
    foo_provider.set(foo)
    assert foo_provider.get() == foo

    # Make sure classes are translated properly too
    BlockingFooProvider.cls_in()
    assert BlockingFooProvider.cls_out() == BlockingFooProvider

    # Make sure the constructor works
    foo = BlockingFoo()
    foo_provider = BlockingFooProvider(foo)
    assert foo_provider.get() == foo



# compatibility utilities/polyfills for supporting older python versions
import importlib
import logging
import sys
import typing

logger = logging.getLogger("synchronicity")

# Modules that cannot be evaluated at runtime, e.g.,
# only available under the TYPE_CHECKING guard, but can be used freely in stub files
TYPE_CHECKING_OVERRIDES = {"_typeshed"}


def evaluated_annotation(annotation, *, globals_=None, declaration_module=None):
    # evaluate string annotations...
    imported_declaration_module = None
    if globals_ is None and declaration_module is not None:
        if declaration_module in sys.modules:
            # already loaded module
            imported_declaration_module = sys.modules[declaration_module]
        else:
            imported_declaration_module = importlib.import_module(declaration_module)
        globals_ = imported_declaration_module.__dict__

    try:
        return eval(annotation, globals_)
    except NameError:
        if "." in annotation:
            # in case of unimported modules referenced in the annotation itself
            # typically happens with TYPE_CHECKING guards etc.
            ref_module, _ = annotation.rsplit(".", 1)
            # for modules that can't be evaluated at runtime,
            # return a ForwardRef with __forward_module__ set
            # to the name of the module that we want to import in the stub file
            if ref_module in TYPE_CHECKING_OVERRIDES:
                ref = typing.ForwardRef(annotation)
                ref.__forward_module__ = ref_module
                return ref
            # hack: import the library *into* the namespace of the supplied globals
            exec(f"import {ref_module}", globals_)
            return eval(annotation, globals_)
        raise



import inspect

from synchronicity import Synchronizer


def f(x):
    return x**2


def test_multiwrap_warning(recwarn):
    s = Synchronizer(multiwrap_warning=True)
    try:
        f_s = s.create_blocking(f)
        assert f_s(42) == 1764
        assert len(recwarn) == 0
        f_s_s = s.create_blocking(f_s)
        assert f_s_s(42) == 1764
        assert len(recwarn) == 1
    finally:
        s._close_loop()  # clean up


def test_multiwrap_no_warning(recwarn, synchronizer):
    f_s = synchronizer.create_blocking(f)
    assert f_s(42) == 1764
    f_s_s = synchronizer.create_blocking(f_s)
    assert f_s_s(42) == 1764
    print("Recorded warnings 1:")
    for w in recwarn.list:
        print(str(w))
    assert len(recwarn) == 0


async def asyncgen():
    yield 42


async def returns_asyncgen():
    return asyncgen()


def test_check_double_wrapped(recwarn, synchronizer):
    assert len(recwarn) == 0
    ret = synchronizer.create_blocking(returns_asyncgen)()
    assert inspect.isasyncgen(ret)
    for w in recwarn.list:
        print("Recorded warning 2:", w)

    assert len(recwarn) == 1



import collections.abc
import contextlib
import functools
import inspect
import typing
from contextlib import asynccontextmanager as _asynccontextmanager

import typing_extensions

from .exceptions import UserCodeException
from .interface import Interface


def wraps_by_interface(interface: Interface, func):
    """Like functools.wraps but maintains `inspect.iscoroutinefunction` and allows custom type annotations overrides

    Use this when the wrapper function is non-async but returns the coroutine resulting
    from calling the underlying wrapped `func`. This will make sure that the wrapper
    is still an async function in that case, and can be inspected as such.

    Note: Does not forward async generator information other than explicit annotations
    """
    if is_coroutine_function_follow_wrapped(func) and interface == Interface._ASYNC_WITH_BLOCKING_TYPES:

        def asyncfunc_deco(user_wrapper):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                try:
                    return await user_wrapper(*args, **kwargs)
                except UserCodeException as uc_exc:
                    uc_exc.exc.__suppress_context__ = True
                    raise uc_exc.exc

            return wrapper

        return asyncfunc_deco
    else:
        return functools.wraps(func)


def is_coroutine_function_follow_wrapped(func: typing.Callable) -> bool:
    """Determine if func returns a coroutine, unwrapping decorators, but not the async synchronicity interace."""
    from .synchronizer import TARGET_INTERFACE_ATTR  # Avoid circular import

    if hasattr(func, "__wrapped__") and getattr(func, TARGET_INTERFACE_ATTR, None) != Interface.BLOCKING:
        return is_coroutine_function_follow_wrapped(func.__wrapped__)
    return inspect.iscoroutinefunction(func)


def is_async_gen_function_follow_wrapped(func: typing.Callable) -> bool:
    """Determine if func returns an async generator, unwrapping decorators, but not the async synchronicity interace."""
    from .synchronizer import TARGET_INTERFACE_ATTR  # Avoid circular import

    if hasattr(func, "__wrapped__") and getattr(func, TARGET_INTERFACE_ATTR, None) != Interface.BLOCKING:
        return is_async_gen_function_follow_wrapped(func.__wrapped__)
    return inspect.isasyncgenfunction(func)


YIELD_TYPE = typing.TypeVar("YIELD_TYPE")
SEND_TYPE = typing.TypeVar("SEND_TYPE")


P = typing_extensions.ParamSpec("P")


def asynccontextmanager(
    f: typing.Callable[P, typing.AsyncGenerator[YIELD_TYPE, SEND_TYPE]],
) -> typing.Callable[P, typing.AsyncContextManager[YIELD_TYPE]]:
    """Wrapper around contextlib.asynccontextmanager that sets correct type annotations

    The standard library one doesn't
    """
    acm_factory: typing.Callable[..., typing.AsyncContextManager[YIELD_TYPE]] = _asynccontextmanager(f)

    old_ret = acm_factory.__annotations__.pop("return", None)
    if old_ret is not None:
        if old_ret.__origin__ in [
            collections.abc.AsyncGenerator,
            collections.abc.AsyncIterator,
            collections.abc.AsyncIterable,
        ]:
            acm_factory.__annotations__["return"] = typing.AsyncContextManager[old_ret.__args__[0]]  # type: ignore
        elif old_ret.__origin__ == contextlib.AbstractAsyncContextManager:
            # if the standard lib fixes the annotations in the future, lets not break it...
            return acm_factory
    else:
        raise ValueError(
            "To use the fixed @asynccontextmanager, make sure to properly"
            " annotate your wrapped function as an AsyncGenerator"
        )

    return acm_factory



"""
Tests exceptions thrown from functions wrapped by Synchronicity.

Currently, exceptions are thrown from Synchronicity like so:

try:
    return self._func(*args, **kwargs)
except UserCodeException as uc_exc:
    uc_exc.exc.__suppress_context__ = True
    raise uc_exc.exc

When we raise an exception, the exception context is from Synchronicity, which
may confuse users. Therefore, we set __suppress_context__ to True to avoid
showing the user those error messages. This will preserve uc_exc.exc.__cause__,
but will cause uc_exc.exc.__context__ to be lost. Unfortunately, I don't know
how to avoid that.

These tests ensure that the __cause__ of an user exception is not lost, and
that either __suppress_context__ is True or __context__ is None so that users
are not exposed to confusing Synchronicity error messages.

See https://github.com/modal-labs/synchronicity/pull/165 for more details.
"""

import asyncio
import concurrent
import functools
import inspect
import pytest
import sys
import time
import typing

SLEEP_DELAY = 0.1
WINDOWS_TIME_RESOLUTION_FIX = 0.01 if sys.platform == "win32" else 0.0


class CustomExceptionCause(Exception):
    pass


class CustomException(Exception):
    pass


async def f_raises():
    await asyncio.sleep(SLEEP_DELAY)
    raise CustomException("something failed")


async def f_raises_with_cause():
    await asyncio.sleep(SLEEP_DELAY)
    raise CustomException("something failed") from CustomExceptionCause("exception cause")


def test_function_raises_sync(synchronizer):
    t0 = time.monotonic()
    with pytest.raises(CustomException) as exc:
        f_raises_s = synchronizer.create_blocking(f_raises)
        f_raises_s()
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= time.monotonic() - t0 < 2 * SLEEP_DELAY
    assert exc.value.__suppress_context__ or exc.value.__context__ is None


def test_function_raises_with_cause_sync(synchronizer):
    t0 = time.monotonic()
    with pytest.raises(CustomException) as exc:
        f_raises_s = synchronizer.create_blocking(f_raises_with_cause)
        f_raises_s()
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= time.monotonic() - t0 < 2 * SLEEP_DELAY
    assert isinstance(exc.value.__cause__, CustomExceptionCause)


def test_function_raises_sync_futures(synchronizer):
    t0 = time.monotonic()
    f_raises_s = synchronizer.create_blocking(f_raises)
    fut = f_raises_s(_future=True)
    assert isinstance(fut, concurrent.futures.Future)
    assert time.monotonic() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException) as exc:
        fut.result()
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= time.monotonic() - t0 < 2 * SLEEP_DELAY
    assert exc.value.__suppress_context__ or exc.value.__context__ is None


def test_function_raises_with_cause_sync_futures(synchronizer):
    t0 = time.monotonic()
    f_raises_s = synchronizer.create_blocking(f_raises_with_cause)
    fut = f_raises_s(_future=True)
    assert isinstance(fut, concurrent.futures.Future)
    assert time.monotonic() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException) as exc:
        fut.result()
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= time.monotonic() - t0 < 2 * SLEEP_DELAY
    assert isinstance(exc.value.__cause__, CustomExceptionCause)


@pytest.mark.asyncio
async def test_function_raises_async(synchronizer):
    t0 = time.monotonic()
    f_raises_s = synchronizer.create_blocking(f_raises)
    coro = f_raises_s.aio()
    assert inspect.iscoroutine(coro)
    assert time.monotonic() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException) as exc:
        await coro
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= time.monotonic() - t0 < 2 * SLEEP_DELAY
    assert exc.value.__suppress_context__ or exc.value.__context__ is None


@pytest.mark.asyncio
async def test_function_raises_with_cause_async(synchronizer):
    t0 = time.monotonic()
    f_raises_s = synchronizer.create_blocking(f_raises_with_cause)
    coro = f_raises_s.aio()
    assert inspect.iscoroutine(coro)
    assert time.monotonic() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException) as exc:
        await coro
    dur = time.monotonic() - t0
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= dur < 2 * SLEEP_DELAY
    assert isinstance(exc.value.__cause__, CustomExceptionCause)


async def f_raises_baseexc():
    await asyncio.sleep(SLEEP_DELAY)
    raise KeyboardInterrupt


def test_function_raises_baseexc_sync(synchronizer):
    t0 = time.monotonic()
    with pytest.raises(BaseException) as exc:
        f_raises_baseexc_s = synchronizer.create_blocking(f_raises_baseexc)
        f_raises_baseexc_s()
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= time.monotonic() - t0 < 2 * SLEEP_DELAY
    assert exc.value.__suppress_context__ or exc.value.__context__ is None


def f_raises_syncwrap() -> typing.Coroutine[typing.Any, typing.Any, None]:
    return f_raises()  # returns a coro


@pytest.mark.asyncio
async def test_function_raises_async_syncwrap(synchronizer):
    t0 = time.monotonic()
    f_raises_syncwrap_s = synchronizer.create_blocking(f_raises_syncwrap)
    coro = f_raises_syncwrap_s.aio()
    assert inspect.iscoroutine(coro)
    assert time.monotonic() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException) as exc:
        await coro
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= time.monotonic() - t0 < 2 * SLEEP_DELAY
    assert exc.value.__suppress_context__ or exc.value.__context__ is None


def f_raises_with_cause_syncwrap() -> typing.Coroutine[typing.Any, typing.Any, None]:
    return f_raises_with_cause()  # returns a coro


@pytest.mark.asyncio
async def test_function_raises_with_cause_async_syncwrap(synchronizer):
    t0 = time.monotonic()
    f_raises_syncwrap_s = synchronizer.create_blocking(f_raises_with_cause_syncwrap)
    coro = f_raises_syncwrap_s.aio()
    assert inspect.iscoroutine(coro)
    assert time.monotonic() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException) as exc:
        await coro
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= time.monotonic() - t0 < 2 * SLEEP_DELAY
    assert isinstance(exc.value.__cause__, CustomExceptionCause)


def decorator(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


f_raises_wrapped = decorator(f_raises)


@pytest.mark.asyncio
async def test_wrapped_function_raises_async(synchronizer):
    t0 = time.monotonic()
    f_raises_s = synchronizer.create_blocking(f_raises_wrapped)
    coro = f_raises_s.aio()
    assert inspect.iscoroutine(coro)
    assert time.monotonic() - t0 < SLEEP_DELAY
    with pytest.raises(CustomException) as exc:
        await coro
    assert SLEEP_DELAY - WINDOWS_TIME_RESOLUTION_FIX <= time.monotonic() - t0 < 2 * SLEEP_DELAY
    assert exc.value.__suppress_context__ or exc.value.__context__ is None



import functools
import typing

import typing_extensions

from synchronicity.async_wrap import wraps_by_interface
from synchronicity.exceptions import UserCodeException
from synchronicity.interface import Interface

if typing.TYPE_CHECKING:
    from synchronicity.synchronizer import Synchronizer


class FunctionWithAio:
    def __init__(self, func, aio_func, synchronizer):
        self._func = func
        self._aio_func = self.aio = aio_func
        self._synchronizer = synchronizer

    def __call__(self, *args, **kwargs):
        # .__call__ is special - it's being looked up on the class instead of the instance when calling something,
        # so setting the magic method from the constructor is not possible
        # https://stackoverflow.com/questions/22390532/object-is-not-callable-after-adding-call-method-to-instance
        # so we need to use an explicit wrapper function here
        try:
            return self._func(*args, **kwargs)
        except UserCodeException as uc_exc:
            uc_exc.exc.__suppress_context__ = True
            raise uc_exc.exc


class MethodWithAio:
    """Creates a bound method that can have callable child-properties on the method itself.

    Child-properties are also bound to the parent instance.
    """

    def __init__(self, func, aio_func, synchronizer: "Synchronizer", is_classmethod=False):
        self._func = func
        self._aio_func = aio_func
        self._synchronizer = synchronizer
        self._is_classmethod = is_classmethod

    def __get__(self, instance, owner=None):
        bind_var = instance if instance is not None and not self._is_classmethod else owner

        bound_func = functools.wraps(self._func)(functools.partial(self._func, bind_var))  # bound blocking function
        self._synchronizer._update_wrapper(bound_func, self._func, interface=Interface.BLOCKING)

        bound_aio_func = wraps_by_interface(Interface._ASYNC_WITH_BLOCKING_TYPES, self._aio_func)(
            functools.partial(self._aio_func, bind_var)
        )  # bound async function
        self._synchronizer._update_wrapper(bound_func, self._func, interface=Interface._ASYNC_WITH_BLOCKING_TYPES)
        bound_func.aio = bound_aio_func
        return bound_func


CTX = typing.TypeVar("CTX", covariant=True)


class AsyncAndBlockingContextManager(typing_extensions.Protocol[CTX]):
    def __enter__(self) -> CTX: ...

    async def __aenter__(self) -> CTX: ...

    def __exit__(self, typ, value, tb): ...

    async def __aexit__(self, typ, value, tb): ...



import asyncio
import pytest
from typing import Any, Dict


@pytest.mark.asyncio
async def test_getattr(synchronizer):
    class Foo:
        _attrs: Dict[str, Any]

        def __init__(self):
            self._attrs = {}

        async def __getattr__(self, k):
            await asyncio.sleep(0.01)
            return self._attrs[k]

        def __setattr__(self, k, v):
            if k in self.__annotations__:
                # Only needed because the constructor sets _attrs
                self.__dict__[k] = v
            else:
                self._attrs[k] = v

        @property
        def z(self):
            return self._attrs["x"]

        @staticmethod
        def make_foo():
            return Foo()

    # original type tests:
    foo = Foo()
    foo.x = 42
    assert await foo.x == 42
    with pytest.raises(KeyError):
        await foo.y
    assert foo.z == 42

    BlockingFoo = synchronizer.create_blocking(Foo)

    blocking_foo = BlockingFoo()
    blocking_foo.x = 43
    assert blocking_foo.x == 43
    with pytest.raises(KeyError):
        blocking_foo.y
    assert blocking_foo.z == 43

    blocking_foo = BlockingFoo.make_foo()
    blocking_foo.x = 44
    assert isinstance(blocking_foo, BlockingFoo)

    # TODO: there is no longer a way to make async properties, but there is this w/ async __getattr__:
    assert await blocking_foo.__getattr__.aio("x") == 44



import inspect
import typing

from synchronicity import async_wrap
from synchronicity.async_wrap import wraps_by_interface
from synchronicity.interface import Interface
from synchronicity.synchronizer import FunctionWithAio


def test_wrap_corofunc_using_async():
    async def foo():
        pass

    @wraps_by_interface(Interface._ASYNC_WITH_BLOCKING_TYPES, foo)
    async def bar():
        pass

    assert inspect.iscoroutinefunction(bar)


def test_wrap_corofunc_using_non_async():
    async def foo():
        pass

    @wraps_by_interface(Interface._ASYNC_WITH_BLOCKING_TYPES, foo)
    def bar():
        pass

    assert inspect.iscoroutinefunction(bar)


def test_wrap_asynccontextmanager_annotations():
    @async_wrap.asynccontextmanager  # this would not work with contextlib.asynccontextmanager
    async def foo() -> typing.AsyncGenerator[int, None]: ...

    assert foo.__annotations__["return"] == typing.AsyncContextManager[int]


def test_wrap_staticmethod(synchronizer):
    class Foo:
        @staticmethod
        async def a_static_method() -> typing.Awaitable[str]:
            async def wrapped():
                return "hello"

            return wrapped()

    BlockingFoo = synchronizer.create_blocking(Foo)

    assert isinstance(BlockingFoo.__dict__["a_static_method"], FunctionWithAio)
    assert not inspect.iscoroutinefunction(BlockingFoo.__dict__["a_static_method"]._func)
    assert inspect.iscoroutinefunction(BlockingFoo.__dict__["a_static_method"].aio)



import asyncio
import pytest

events = []


async def async_producer():
    for i in range(10):
        events.append("producer")
        yield i


@pytest.mark.asyncio
async def test_generator_order_async(synchronizer):
    events.clear()
    async_producer_synchronized = synchronizer.create_blocking(async_producer)
    async for i in async_producer_synchronized.aio():
        events.append("consumer")
    assert events == ["producer", "consumer"] * 10


@pytest.mark.asyncio
async def test_generator_order_explicit_async(synchronizer):
    events.clear()
    async_producer_synchronized = synchronizer.create_blocking(async_producer)
    async for i in async_producer_synchronized.aio():
        events.append("consumer")
    assert events == ["producer", "consumer"] * 10


def test_generator_order_sync(synchronizer):
    events.clear()
    async_producer_synchronized = synchronizer.create_blocking(async_producer)
    for i in async_producer_synchronized():
        events.append("consumer")
    assert events == ["producer", "consumer"] * 10


async def async_bidirectional_producer(i):
    j = yield i
    assert j == i**2


@pytest.mark.asyncio
async def test_bidirectional_generator_async(synchronizer):
    f = synchronizer.create_blocking(async_bidirectional_producer)
    gen = f.aio(42)
    value = await gen.asend(None)
    assert value == 42
    with pytest.raises(StopAsyncIteration):
        await gen.asend(42 * 42)


def test_bidirectional_generator_sync(synchronizer):
    f = synchronizer.create_blocking(async_bidirectional_producer)
    gen = f(42)
    value = gen.send(None)
    assert value == 42
    with pytest.raises(StopIteration):
        gen.send(42 * 42)


async def athrow_example_gen():
    try:
        await asyncio.sleep(0.1)
        yield "hello"
    except ZeroDivisionError:
        await asyncio.sleep(0.2)
        yield "world"
    except BaseException:
        yield "foobar"


@pytest.mark.asyncio
async def test_athrow_async(synchronizer):
    gen = synchronizer.create_blocking(athrow_example_gen).aio()
    v = await gen.asend(None)
    assert v == "hello"
    v = await gen.athrow(ZeroDivisionError)
    assert v == "world"
    await gen.aclose()


def test_athrow_sync(synchronizer):
    gen = synchronizer.create_blocking(athrow_example_gen)()
    v = gen.send(None)
    assert v == "hello"
    v = gen.throw(ZeroDivisionError)
    assert v == "world"


@pytest.mark.asyncio
async def test_athrow_baseexc_async(synchronizer):
    gen = synchronizer.create_blocking(athrow_example_gen).aio()
    v = await gen.asend(None)
    assert v == "hello"
    v = await gen.athrow(KeyboardInterrupt)
    assert v == "foobar"


def test_athrow_baseexc_sync(synchronizer):
    gen = synchronizer.create_blocking(athrow_example_gen)()
    v = gen.send(None)
    assert v == "hello"
    v = gen.throw(KeyboardInterrupt)
    assert v == "foobar"


async def ensure_stop_async_iteration():
    try:
        yield 42
        yield 43
    except BaseException as exc:
        events.append(exc)


def test_ensure_stop_async_iteration(synchronizer):
    events.clear()

    def create_generator():
        gen_f = synchronizer.create_blocking(ensure_stop_async_iteration)
        for x in gen_f():
            break

    create_generator()
    assert len(events) == 1
    assert isinstance(events[0], GeneratorExit)


class MyGenerator:
    def __aiter__(self):
        return async_producer()


def test_custom_generator(synchronizer):
    events.clear()
    BlockingMyGenerator = synchronizer.create_blocking(MyGenerator)
    blocking_my_generator = BlockingMyGenerator()
    for x in blocking_my_generator:
        events.append("consumer")
    assert events == ["producer", "consumer"] * 10



# this code is only meant to be "running" through mypy and not an actual python interpreter!
import typing

from typing_extensions import assert_type

from test.type_stub_helpers import e2e_example_export

blocking_foo = e2e_example_export.BlockingFoo("hello")

e2e_example_export.listify(123)


import asyncio
import sys
import time
from contextlib import asynccontextmanager

from synchronicity import Synchronizer


@asynccontextmanager
async def ctx_mgr():
    try:
        if sys.argv[1] == "enter":
            while True:
                print("enter")
                await asyncio.sleep(0.3)
        elif sys.argv[1] == "yield":
            yield
        else:
            print("this should not happen")
    finally:
        print("exit")


s = Synchronizer()
blocking_ctx_mgr = s.create_blocking(ctx_mgr)
try:
    with blocking_ctx_mgr():
        while True:
            print("in ctx")
            time.sleep(0.3)
except KeyboardInterrupt:
    print("keyboard interrupt")



import typing
from typing import AsyncGenerator, List, TypeVar, Union, overload

import typing_extensions

from synchronicity.async_wrap import asynccontextmanager


class _Foo:
    singleton: "_Foo"

    def __init__(self, arg: str):
        self.arg = arg

    async def getarg(self) -> str:
        return self.arg

    async def gen(self) -> AsyncGenerator[int, None]:
        yield 1

    @staticmethod
    def some_static(arg: str) -> float:
        return  # type: ignore

    @classmethod
    def clone(cls, foo: "_Foo") -> "_Foo":  # self ref
        return  # type: ignore


_T = TypeVar("_T", bound=_Foo)


async def _listify(t: _T) -> List[_T]:
    return [t]


@overload
def _overloaded(arg: str) -> float:
    pass


@overload
def _overloaded(arg: int) -> int:
    pass


def _overloaded(arg: Union[str, int]):
    if isinstance(arg, str):
        return float(arg)
    return arg


async def _returns_foo() -> _Foo:
    return _Foo("hello")


@asynccontextmanager
async def make_context(a: float) -> typing.AsyncGenerator[str, None]:
    yield "hello"


P = typing_extensions.ParamSpec("P")
R = typing.TypeVar("R")


class CallableWrapper(typing.Generic[P, R]):
    async def func(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return  # type: ignore


def wrap_callable(c: typing.Callable[P, R]) -> CallableWrapper[P, R]:
    return  # type: ignore



import asyncio
import pytest
import typing

from synchronicity import Synchronizer


class DummyConnection:
    async def run_query(self, query):
        pass


async def dummy_connect_to_db(url):
    return DummyConnection()


@pytest.fixture()
def quicksleep(monkeypatch):
    from asyncio import sleep as original_sleep

    monkeypatch.setattr("asyncio.sleep", lambda x: original_sleep(x / 1000.0))


def pytest_markdown_docs_globals():
    synchronizer = Synchronizer()
    return {
        "typing": typing,
        "synchronizer": synchronizer,
        "asyncio": asyncio,
        "connect_to_database": dummy_connect_to_db,
    }



def test_is_synchronized(synchronizer):
    class Foo:
        pass

    BlockingFoo = synchronizer.create_blocking(Foo)
    assert synchronizer.is_synchronized(Foo) is False
    assert synchronizer.is_synchronized(BlockingFoo) is True



import pytest
import subprocess
import sys
from pathlib import Path


@pytest.mark.skipif(sys.version_info >= (3, 13), reason="gevent seems broken on Python 3.13")
def test_gevent():
    # Run it in a separate process because gevent modifies a lot of modules
    fn = Path(__file__).parent / "support" / "_gevent.py"
    ret = subprocess.run([sys.executable, fn], stdout=sys.stdout, stderr=sys.stderr, timeout=5)
    assert ret.returncode == 0



import pytest
import subprocess
import sys
from pathlib import Path


@pytest.mark.skipif(sys.platform == "win32", reason="Windows can't fork")
def test_fork_restarts_loop():
    with subprocess.Popen(
        [sys.executable, Path(__file__).parent / "support" / "_forker.py"],
        encoding="utf8",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as p:
        try:
            stdout, stderr = p.communicate(timeout=2)
        except subprocess.TimeoutExpired:
            p.kill()
            assert False, "Fork process hanged"

        assert p.returncode == 0
        assert stdout == "done\ndone\n"



import typing
import typing_extensions

class __foo_spec(typing_extensions.Protocol):
    def __call__(self) -> typing.Generator[int, None, None]:
        ...

    def aio(self) -> typing.AsyncGenerator[int, None]:
        ...

foo: __foo_spec



import pytest
import signal
import subprocess
import sys
from pathlib import Path


class PopenWithCtrlC(subprocess.Popen):
    def __init__(self, *args, creationflags=0, **kwargs):
        if sys.platform == "win32":
            # needed on windows to separate ctrl-c lifecycle of subprocess from parent:
            creationflags = creationflags | subprocess.CREATE_NEW_CONSOLE  # type: ignore

        super().__init__(*args, **kwargs, creationflags=creationflags)

    def send_ctrl_c(self):
        # platform independent way to replicate the behavior of Ctrl-C:ing a cli app
        if sys.platform == "win32":
            # windows doesn't support sigint, and subprocess.CTRL_C_EVENT has a bunch
            # of gotchas since it's bound to a console which is the same for the parent
            # process by default, and can't be sent using the python standard library
            # to a separate process's console
            import console_ctrl

            console_ctrl.send_ctrl_c(self.pid)  # noqa [E731]
        else:
            self.send_signal(signal.SIGINT)


def test_shutdown():
    # We run it in a separate process so we can simulate interrupting it
    fn = Path(__file__).parent / "support" / "_shutdown.py"
    p = PopenWithCtrlC([sys.executable, "-u", fn], stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf8")
    for i in range(2):  # this number doesn't matter, it's a while loop
        assert p.stdout.readline() == "running\n"
    p.send_ctrl_c()
    for i in range(2):
        # in some extreme cases there is a risk of a race where the "running" still appears here
        if p.stdout.readline() == "cancelled\n":
            break
    else:
        assert False

    assert p.stdout.readline() == "handled cancellation\n"
    assert p.stdout.readline() == "exit async\n"
    assert (
        p.stdout.readline() == "keyboard interrupt\n"
    )  # we want the keyboard interrupt to come *after* the running function has been cancelled!

    stderr_content = p.stderr.read()
    print("stderr:", stderr_content)
    assert "Traceback" not in stderr_content


def test_keyboard_interrupt_reraised_as_is(synchronizer):
    @synchronizer.create_blocking
    async def a():
        raise KeyboardInterrupt()

    with pytest.raises(KeyboardInterrupt):
        a()


def test_shutdown_during_ctx_mgr_setup():
    # We run it in a separate process so we can simulate interrupting it
    fn = Path(__file__).parent / "support" / "_shutdown_ctx_mgr.py"
    p = PopenWithCtrlC(
        [sys.executable, "-u", fn, "enter"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf8",
    )
    for i in range(2):  # this number doesn't matter, it's a while loop
        assert p.stdout.readline() == "enter\n"
    p.send_ctrl_c()
    assert p.stdout.readline() == "exit\n"
    assert p.stdout.readline() == "keyboard interrupt\n"
    assert p.stderr.read() == ""


def test_shutdown_during_ctx_mgr_yield():
    # We run it in a separate process so we can simulate interrupting it
    fn = Path(__file__).parent / "support" / "_shutdown_ctx_mgr.py"
    p = PopenWithCtrlC(
        [sys.executable, "-u", fn, "yield"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf8"
    )
    for i in range(2):  # this number doesn't matter, it's a while loop
        assert p.stdout.readline() == "in ctx\n"
    p.send_ctrl_c()
    assert p.stdout.readline() == "exit\n"
    assert p.stdout.readline() == "keyboard interrupt\n"
    assert p.stderr.read() == ""


@pytest.mark.parametrize("run_number", range(10))  # don't allow this to flake!
def test_shutdown_during_async_run(run_number):
    fn = Path(__file__).parent / "support" / "_shutdown_async_run.py"
    p = PopenWithCtrlC(
        [sys.executable, "-u", fn],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )

    def line():
        # debugging help
        line_data = p.stdout.readline()
        print(line_data)
        return line_data

    assert line() == "running\n"
    p.send_ctrl_c()
    print("sigint sent")
    while (next_line := line()) == "running\n":
        pass
    assert next_line == "cancelled\n"
    stdout, stderr = p.communicate(timeout=5)
    print(stderr)
    assert (
        stdout
        == """handled cancellation
exit async
keyboard interrupt
"""
    )
    assert stderr == ""



import collections.abc
import synchronicity.combined_types
import typing
import typing_extensions

SUPERSELF = typing.TypeVar("SUPERSELF", covariant=True)

class BlockingFoo:

    singleton: BlockingFoo

    def __init__(self, arg: str):
        ...

    class __getarg_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, /) -> str:
            ...

        async def aio(self, /) -> str:
            ...

    getarg: __getarg_spec[typing_extensions.Self]

    class __gen_spec(typing_extensions.Protocol[SUPERSELF]):
        def __call__(self, /) -> typing.Generator[int, None, None]:
            ...

        def aio(self, /) -> typing.AsyncGenerator[int, None]:
            ...

    gen: __gen_spec[typing_extensions.Self]

    @staticmethod
    def some_static(arg: str) -> float:
        ...

    @classmethod
    def clone(cls, foo: BlockingFoo) -> BlockingFoo:
        ...


_T_Blocking = typing.TypeVar("_T_Blocking", bound="BlockingFoo")

class __listify_spec(typing_extensions.Protocol):
    def __call__(self, /, t: _T_Blocking) -> typing.List[_T_Blocking]:
        ...

    async def aio(self, /, t: _T_Blocking) -> typing.List[_T_Blocking]:
        ...

listify: __listify_spec


@typing.overload
def overloaded(arg: str) -> float:
    ...

@typing.overload
def overloaded(arg: int) -> int:
    ...


class __returns_foo_spec(typing_extensions.Protocol):
    def __call__(self, /) -> BlockingFoo:
        ...

    async def aio(self, /) -> BlockingFoo:
        ...

returns_foo: __returns_foo_spec


class __wrapped_make_context_spec(typing_extensions.Protocol):
    def __call__(self, /, a: float) -> synchronicity.combined_types.AsyncAndBlockingContextManager[str]:
        ...

    def aio(self, /, a: float) -> typing.AsyncContextManager[str]:
        ...

wrapped_make_context: __wrapped_make_context_spec


P = typing_extensions.ParamSpec("P")

R = typing.TypeVar("R")

R_INNER = typing.TypeVar("R_INNER", covariant=True)

P_INNER = typing_extensions.ParamSpec("P_INNER")

class CallableWrapper(typing.Generic[P, R]):
    """Abstract base class for generic types.

    A generic type is typically declared by inheriting from
    this class parameterized with one or more type variables.
    For example, a generic mapping type might be defined as::

      class Mapping(Generic[KT, VT]):
          def __getitem__(self, key: KT) -> VT:
              ...
          # Etc.

    This class can then be used as follows::

      def lookup_name(mapping: Mapping[KT, VT], key: KT, default: VT) -> VT:
          try:
              return mapping[key]
          except KeyError:
              return default
    """
    def __init__(self, /, *args, **kwargs):
        """Initialize self.  See help(type(self)) for accurate signature."""
        ...

    class __func_spec(typing_extensions.Protocol[R_INNER, P_INNER, SUPERSELF]):
        def __call__(self, /, *args: P_INNER.args, **kwargs: P_INNER.kwargs) -> R_INNER:
            ...

        async def aio(self, /, *args: P_INNER.args, **kwargs: P_INNER.kwargs) -> R_INNER:
            ...

    func: __func_spec[R, P, typing_extensions.Self]


def wrap_callable(c: collections.abc.Callable[P, R]) -> CallableWrapper[P, R]:
    ...


some_instance: typing.Optional[BlockingFoo]


import os

import synchronicity

synchronizer = synchronicity.Synchronizer()


@synchronizer.create_blocking
async def dummy():
    print("done", flush=True)


if __name__ == "__main__":
    dummy()  # this starts a synchronizer loop/thread
    if not os.fork():
        assert not synchronizer._thread.is_alive()  # threads don't come along in forks
        dummy()  # this should still work



# This file creates wrapped entities for async implementation in e2e_example_impl.py
# This file is then used as input for generating type stubs
from typing import Optional

import synchronicity

from . import e2e_example_impl

synchronizer = synchronicity.Synchronizer()
BlockingFoo = synchronizer.create_blocking(e2e_example_impl._Foo, "BlockingFoo", __name__)

some_instance: Optional[BlockingFoo] = None

_T_Blocking = synchronizer.create_blocking(
    e2e_example_impl._T, "_T_Blocking", __name__
)  # synchronize the TypeVar to support translation of bounds
listify = synchronizer.create_blocking(e2e_example_impl._listify, "listify", __name__)

overloaded = synchronizer.create_blocking(e2e_example_impl._overloaded, "overloaded", __name__)

returns_foo = synchronizer.create_blocking(e2e_example_impl._returns_foo, "returns_foo", __name__)

wrapped_make_context = synchronizer.create_blocking(e2e_example_impl.make_context, "make_context", __name__)

# TODO: we shouldn't need to wrap typevars unless they have wrapped `bounds`
P = synchronizer.create_blocking(e2e_example_impl.P, "P", __name__)
R = synchronizer.create_blocking(e2e_example_impl.R, "R", __name__)


CallableWrapper = synchronizer.create_blocking(e2e_example_impl.CallableWrapper, "CallableWrapper", __name__)

wrap_callable = synchronizer.create_blocking(e2e_example_impl.wrap_callable, "wrap_callable", __name__)



import contextlib
import pytest
import typing

from synchronicity.async_wrap import asynccontextmanager


async def noop():
    pass


async def error():
    raise Exception("problem")


class Resource:
    def __init__(self):
        self.state = "none"

    def get_state(self):
        return self.state

    @asynccontextmanager
    async def wrap(self) -> typing.AsyncGenerator[None, None]:
        self.state = "entered"
        try:
            yield
        finally:
            self.state = "exited"

    @asynccontextmanager
    async def wrap_yield_twice(self) -> typing.AsyncGenerator[None, None]:
        yield
        yield

    @asynccontextmanager
    async def wrap_never_yield(self) -> typing.AsyncGenerator[None, None]:
        if False:
            yield


def test_asynccontextmanager_sync(synchronizer):
    r = synchronizer.create_blocking(Resource)()
    assert r.get_state() == "none"
    with r.wrap():
        assert r.get_state() == "entered"
    assert r.get_state() == "exited"


@pytest.mark.asyncio
async def test_asynccontextmanager_async(synchronizer):
    r = synchronizer.create_blocking(Resource)()
    assert r.get_state() == "none"
    async with r.wrap():
        assert r.get_state() == "entered"
    assert r.get_state() == "exited"


@pytest.mark.asyncio
async def test_asynccontextmanager_async_raise(synchronizer):
    r = synchronizer.create_blocking(Resource)()
    assert r.get_state() == "none"
    with pytest.raises(Exception):
        async with r.wrap():
            assert r.get_state() == "entered"
            raise Exception("boom")
    assert r.get_state() == "exited"


@pytest.mark.asyncio
async def test_asynccontextmanager_yield_twice(synchronizer):
    r = synchronizer.create_blocking(Resource)()
    with pytest.raises(RuntimeError):
        async with r.wrap_yield_twice():
            pass


@pytest.mark.asyncio
async def test_asynccontextmanager_never_yield(synchronizer):
    r = synchronizer.create_blocking(Resource)()
    with pytest.raises(RuntimeError):
        async with r.wrap_never_yield():
            pass


@pytest.mark.asyncio
async def test_asynccontextmanager_nested(synchronizer):
    finally_blocks = []

    @synchronizer.create_blocking
    @asynccontextmanager
    async def a() -> typing.AsyncGenerator[str, None]:
        try:
            yield "foo"
        finally:
            finally_blocks.append("A")

    @synchronizer.create_blocking
    @asynccontextmanager
    async def b() -> typing.AsyncGenerator[str, None]:
        async with a() as it:
            try:
                yield it
            finally:
                finally_blocks.append("B")

    with pytest.raises(BaseException):
        async with b():
            raise BaseException("boom!")

    assert finally_blocks == ["B", "A"]


@pytest.mark.asyncio
async def test_asynccontextmanager_with_in_async(synchronizer):
    r = synchronizer.create_blocking(Resource)()
    # err_cls = AttributeError if sys.version_info < (3, 11) else TypeError
    # with pytest.raises(err_cls):
    with r.wrap.aio():  # TODO: this *should* not be allowed, but works for stupid reasons
        pass


@pytest.mark.asyncio
async def test_returning_context_manager(synchronizer):
    @contextlib.asynccontextmanager
    async def foo():
        yield "hello"

    @synchronizer.wrap
    def returner() -> typing.AsyncContextManager[str]:
        return foo()

    with returner():
        pass

    async with returner.aio():
        pass



import pytest

from synchronicity import Synchronizer


@pytest.fixture()
def synchronizer():
    s = Synchronizer()
    yield s
    s._close_loop()  # avoid "unclosed event loop" warnings in tests when garbage collecting synchronizers



import inspect


class _Api:
    def blocking_func(self):
        pass

    async def async_func(self):
        pass


def test_inspect_coroutinefunction(synchronizer):
    BlockingApi = synchronizer.create_blocking(_Api)

    assert inspect.iscoroutinefunction(BlockingApi.blocking_func) is False
    assert inspect.iscoroutinefunction(BlockingApi.async_func) is False
    assert hasattr(BlockingApi.blocking_func, "aio") is False
    assert inspect.iscoroutinefunction(BlockingApi.async_func.aio) is True



import enum


class Interface(enum.Enum):
    BLOCKING = enum.auto()
    _ASYNC_WITH_BLOCKING_TYPES = enum.auto()  # this is *only* used for functions, since all types are blocking


# Default names for classes
DEFAULT_CLASS_PREFIX = "Blocking"

# Default names for functions
DEFAULT_FUNCTION_PREFIXES = {
    Interface.BLOCKING: "blocking_",
    # this is only used internally - usage will be via `.aio` on the blocking function:
    Interface._ASYNC_WITH_BLOCKING_TYPES: "aio_",
}



import pytest
import typing

from synchronicity import Synchronizer, combined_types
from synchronicity.interface import Interface
from synchronicity.type_stubs import StubEmitter


class ImplType:
    attr: str


synchronizer = Synchronizer()


@pytest.fixture(autouse=True, scope="module")
def synchronizer_teardown():
    yield
    synchronizer._close_loop()  # prevent "unclosed event loop" warnings


BlockingType = synchronizer.create_blocking(ImplType, "BlockingType", __name__)


def test_wrapped_class_keeps_class_annotations():
    assert BlockingType.__annotations__ == ImplType.__annotations__


@pytest.mark.parametrize(
    "t,interface,expected",
    [
        (
            typing.AsyncGenerator[int, str],
            Interface.BLOCKING,
            typing.Generator[int, str, None],
        ),
        (
            typing.AsyncContextManager[ImplType],
            Interface.BLOCKING,
            combined_types.AsyncAndBlockingContextManager[BlockingType],
        ),
        (
            typing.AsyncContextManager[ImplType],
            Interface._ASYNC_WITH_BLOCKING_TYPES,
            typing.AsyncContextManager[BlockingType],
        ),
        (
            typing.Awaitable[typing.Awaitable[str]],
            Interface._ASYNC_WITH_BLOCKING_TYPES,
            typing.Awaitable[typing.Awaitable[str]],
        ),
        (typing.Awaitable[typing.Awaitable[str]], Interface.BLOCKING, str),
        (typing.Coroutine[None, None, str], Interface.BLOCKING, str),
        (typing.AsyncIterable[str], Interface.BLOCKING, typing.Iterable[str]),
        (typing.AsyncIterator[str], Interface.BLOCKING, typing.Iterator[str]),
        (
            typing.Optional[ImplType],
            Interface.BLOCKING,
            typing.Union[BlockingType, None],
        ),
        (typing.Optional[ImplType], Interface._ASYNC_WITH_BLOCKING_TYPES, typing.Union[BlockingType, None]),
    ],
)
def test_annotation_mapping(t, interface, expected):
    stub_emitter = StubEmitter(__name__)
    assert stub_emitter._translate_annotation(t, synchronizer, interface, __name__) == expected
